"""
Benchmarks for 'FastArray' operators (e.g. ``+``, ``*``, ``%``).
"""
__all__ = [
    "bench_op_add",
]

import itertools
import logging
from itertools import product
from typing import List, Tuple

import numpy as np
from numpy.random import default_rng

from ..rt_categorical import Categorical
from ..rt_dataset import Dataset
from ..rt_numpy import empty
from .benchmark import timestamper
from .rand_data import rand_dataset
from .runner import benchmark, create_comparison_dataset, create_trial_dataset

logger = logging.getLogger(__name__)
"""The logger for this module."""


def bench_op_add(**kwargs) -> Dataset:
    # Implement a benchmark that uses the __add__ operator on two FastArrays
    raise NotImplementedError()


# TODO: Implement benchmarks for other operators; for each one, need to try all applicable cases:
#   * ndarray vs. FastArray
#   * scalar OP array
#   * array OP scalar
#   * array OP array
#   * for each of the above, test various combinations of dtypes as well e.g. add an int16 array to an int64 array



import unittest
from math import isclose

import pytest
import riptide_cpp as rc

from riptable import FastArray as FA
from riptable import *
from riptable.rt_enum import (
    INVALID_DICT,
    DisplayColumnColors,
    DisplayJustification,
    DisplayLength,
    NumpyCharTypes,
    gBinaryUFuncs,
)

float_types = [np.float32, np.float64]
int_types = [
    np.bool_,
    np.int8,
    np.uint8,
    np.int16,
    np.uint16,
    np.int32,
    np.uint32,
    np.intc,
    np.intp,
    np.int_,
    np.int64,
    np.uint64,
]
string_types = [np.bytes_, np.str_]
warning_types = [object, np.complex64, np.complex128, np.float16]

# nan_funcs   = [ np.nanargmax, np.nanargmin, np.nancumprod, np.nancumsum, np.nanmax,
#                np.nanmean, np.nanmedian, np.nanmin, np.nanpercentile, np.nanprod,
#                np.nanstd, np.nansum, np.nanvar ]
# nan_results = []

nan_funcs = [nansum, nanmean, nanmin, nanmax]
no_negatives = [np.log, np.log2, np.log10, np.sqrt]

num_list = list(range(10))
str_list = ["a", "b", "c", "d", "e"]

# a set is now valid
invalid_data = [slice(1, 2, 3), ..., {"a": 1}]  # add type, elipses, dict

np.random.seed(1234)


def almost_eq(fa: FastArray, arr: np.ndarray, places=5, rtol=1e-5, atol=1e-5, equal_nan=True):
    epsilon = 10 ** (-places)
    rtol = epsilon

    if fa.shape != arr.shape:
        return False

    fa_arr = fa
    np_arr = arr
    if np.issubdtype(fa_arr.dtype, np.floating):
        # N.B.: Do not compare the elements that both numpy and riptable give nan or inf.
        #       In the case of modulo by zero, numpy gives nan but riptable gives inf.
        finite_mask = np.isfinite(fa_arr) | np.isfinite(np_arr)
        if not np.allclose(fa_arr[finite_mask], np_arr[finite_mask], rtol, atol, equal_nan):
            return False
    else:
        try:
            if not (fa_arr == np_arr).all():
                return False
        except TypeError:
            return False
    return True


class FastArray_Test(unittest.TestCase):

    # --------VALID CONSTRUCTORS-------------------
    """
    BUGS:
    when initializing from a scalar, the requested dtype is ignored
    it will default to the type of the scalar OR the default numpy type for the python type

    Numpy does not ignore the dtype keyword arg.
    """

    def test_float_constructors(self):
        for dt in float_types:
            arr = FastArray(num_list, dtype=dt)
            self.assertEqual(arr.dtype, dt)

        # from scalars
        for dt in float_types:
            arr = FastArray(num_list[0], dtype=dt)
            self.assertEqual(arr.dtype, dt)

    def test_int_constructors(self):
        for dt in int_types:
            arr = FastArray(num_list, dtype=dt)
            self.assertEqual(arr.dtype, dt)
        # from scalars
        for dt in int_types:
            arr = FastArray(num_list[0], dtype=dt)
            self.assertEqual(arr.dtype, dt)

    # --------INVALID CONSTRUCTORS-----------------
    def test_invalid_types(self):
        for dt in warning_types:
            with self.assertWarns(UserWarning):
                _ = FastArray([1], dtype=dt)

    def test_invalid_data(self):
        for d in invalid_data:
            with self.assertRaises(TypeError):
                _ = FastArray(d)

    ## --------TEST SFW INTEGER NAN VALUES-----------
    #'''
    # BUGS:
    # nanstd: mean considers our custom nan values, variance does not
    #'''
    def test_nan_values_int8(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.int8).num]], dtype=np.int8)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_uint8(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.uint8).num]], dtype=np.uint8)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_int16(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.int16).num]], dtype=np.int16)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_uint16(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.uint16).num]], dtype=np.uint16)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_int32(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.int32).num]], dtype=np.int32)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_uint32(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.uint32).num]], dtype=np.uint32)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_int64(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.int64).num]], dtype=np.int64)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    def test_nan_values_uint64(self):
        arr = FastArray([1, INVALID_DICT[np.dtype(np.uint64).num]], dtype=np.uint64)
        for f in nan_funcs:
            result = f(arr)
            self.assertEqual(result, 1)

    # --------OVERFLOW / CASTING TESTS-----------
    def test_overflow_int64(self):
        """
        FastArray flips overflow values in integer types to floats.
        Numpy produces an incorrect result.
        """
        a = 0xFFFFFFFFFFFFFFF
        arr_f = FastArray([a] * 20, dtype=np.int64)
        sum_f = arr_f.sum()

        # The FastArray sum is expected to detect the overflowing integer math,
        # continue accumulating as a float64, then return a float64.
        # Make sure it did.
        self.assertIsInstance(sum_f, np.float64)

        # Calculate the correct result using Python's arbitrary-precision integer support;
        # convert the result to a float64 value for comparison against the
        # sum returned by riptable.
        int_sum_f = a * 20
        self.assertEqual(sum_f, np.float64(int_sum_f))

    def test_overflow_uint64(self):
        """
        FastArray flips overflow values in integer types to floats.
        Numpy produces an incorrect result.
        """
        a = 0xFFFFFFFFFFFFFFF
        arr_f = FastArray([a] * 20, dtype=np.uint64)
        sum_f = arr_f.sum()

        # The FastArray sum is expected to detect the overflowing integer math,
        # continue accumulating as a float64, then return a float64.
        # Make sure it did.
        self.assertIsInstance(sum_f, np.float64)

        # Calculate the correct result using Python's arbitrary-precision integer support;
        # convert the result to a float64 value for comparison against the
        # sum returned by riptable.
        int_sum_f = a * 20
        self.assertEqual(sum_f, np.float64(int_sum_f))

    # ----------INPLACE OPERATIONS----------
    def test_inplace_int_float(self):
        """
        Unlike numpy arrays, FastArray allows inplace operations between integer arrays and floating point scalars.
        The datatype of the array will remain the same. However currently division is not supported.

        Potential BUG: The floor division operator does not raise an error.
        """
        nums = [1, 2, 3, 4, 5]
        for dt in int_types[1:]:  # dont include bool
            arr = FA(nums, dtype=dt)
            for dtf in float_types:
                scalar = dtf(1)
                arr += scalar
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after += operation",
                )
                arr -= scalar
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after -= operation",
                )
                arr *= scalar
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after *= operation",
                )
                arr /= scalar
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after /= opration",
                )
                arr //= scalar
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after //= operation",
                )

        for dt in int_types[1:]:  # dont include bool
            arr = FA(nums, dtype=dt)
            for dtf in float_types:
                arr2 = arr.astype(dtf)
                arr += arr2
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after += operation",
                )
                arr -= arr2
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after -= operation",
                )
                arr *= arr2
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after *= operation",
                )
                arr /= arr2
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after /= operation",
                )
                arr //= arr2
                self.assertEqual(
                    arr.dtype,
                    dt,
                    msg=f"Result datatype {arr.dtype} did not match original datatype {dt} after //= operation",
                )

    def test_forced_dtype_binary(self):
        """
        Test to see that datatype is being correctly forced for result of binary ufuncs.
        You cannot force the type to be an unsigned integer.
        You cannot force the type for a division operation.
        """
        nums = [1, 2, 3, 4, 5]
        types = int_types[1:] + float_types
        types = [i for i in types if i.__name__[0] != "u"]  # remove unsigned integers
        binary_funcs = [
            np.add,
            np.subtract,
            np.multiply,
        ]  # , np.divide, np.floor_divide ]
        for dt1 in types:
            arr1 = FA(nums, dtype=dt1)
            for dt2 in types:
                arr2 = FA(nums, dtype=dt2)
                for dt_forced in types:
                    for func in binary_funcs:
                        result = func(arr1, arr2, dtype=dt_forced)
                        self.assertEqual(
                            result.dtype,
                            dt_forced,
                            msg=f"Unable to force dtype {dt_forced} for binary func {func} between {dt1} and {dt2}",
                        )

    # def test_forced_dtype_unary(self):
    #    nums = [1,2,3,4,5]
    #    unary_funcs = [ np.sum, np.mean, np.std, np.var,
    #                    np.nansum, np.nanmean, np.nanstd, np.nanvar ]
    #    for dt in int_types[1:] + float_types:
    #        arr = FA(nums, dtype=dt)
    #        for dt_forced in int_types[1:] + float_types:
    #            for func in unary_funcs:
    #                result = func(arr, dtype=dt_forced)
    #                self.assertEqual(type(result), dt_forced, msg=f"Unable to force dtype {dt_forced} for unary func {func} on array of type {dt}")

    # def test_forced_dtype_unary(self):
    #    unary_funcs = [ np.absolute, np.abs, np.fabs, np.floor, np.ceil, np.trunc, np.round, np.rint ]
    #    nums = [1,2,3,4,5]
    #    for dt in int_types[1:] + float_types:
    #        arr = FA(nums, dtype=dt)
    #        for dt_forced in int_types[1:] + float_types:
    #            for func in unary_funcs:
    #                try:
    #                    result = func(arr, dtype=dt_forced)
    #                except:
    #                    print(f"unable to perform {func} on {dt}")
    #                #self.assertEqual(result.dtype, dt_forced, msg=f"Result datatype {result.dtype} did not match requested datatype {dt_forced} for input datatype {dt} for {func} operation")

    # --------REDUCE FUNCTION CASTING-----------
    """
    For certain reduce functions on certain types, FastArray will return a different datatype than numpy
    """

    def test_reduce_sums(self):
        reduce_funcs = [np.sum, np.nansum]
        usi64 = NumpyCharTypes.UnsignedInteger64
        si64 = NumpyCharTypes.SignedInteger64
        f64 = NumpyCharTypes.Float64
        input_types = [
            np.bool_,
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
            np.intc,
            np.intp,
            np.int_,
            np.float32,
            np.float64,
        ]
        correct_list = [
            si64,
            si64,
            si64,
            si64,
            si64,
            si64,
            si64,
            si64,
            usi64,
            si64,
            si64,
            si64,
            f64,
            f64,
        ]
        correct_dict = dict(zip(input_types, correct_list))

        arr = FastArray(num_list)
        for dt in input_types:
            arr = arr.astype(dt)
            for f in reduce_funcs:
                reduce_dt = f(arr).dtype
                reduce_dt_char = reduce_dt.char
                self.assertIn(
                    reduce_dt_char,
                    correct_dict[dt],
                    msg=f"Reduce function {f} returned the wrong dtype {reduce_dt_char} ({reduce_dt}) for {dt} input.",
                )

    # -------------- REDUCE FUNCTION ACCURACY ---------
    def test_reduce_accuracy(self):
        """
        FastArray often performs math operations at a higher level of precision than numpy. As a
        result, the answers may be slightly different.
        """
        reduce_funcs = [
            np.sum,
            np.mean,
            np.std,
            np.var,
            np.nansum,
            np.nanmean,
            np.nanstd,
            np.nanvar,
        ]
        # use a large array to illustrate the differences
        num_list = (np.random.rand(1000000) * 100) + 1
        num_types = int_types + float_types

        for dt in num_types:
            fa_arr = FA(num_list, dtype=dt)
            np_arr = np.array(num_list, dtype=dt)
            for func in reduce_funcs:
                fa_result = func(fa_arr)
                np_result = func(np_arr)
                # 4 places of accuracy seems to solve differences in float32/64
                # TODO: Change to use self.assertAlmostEqual() instead of isclose?
                is_accurate = isclose(fa_result, np_result, rel_tol=1e-06)
                self.assertTrue(
                    is_accurate,
                    msg=f"Reduce function {func} produced fa: {fa_result} np: {np_result} for datatype {dt} with data: {fa_arr}",
                )

    # -------REDUCE FUNCTION PRESERVING---------
    def test_preserve_reduce_type(self):
        """
        Certain reduce functions need to preserve their type. This test is to make sure FastArray
        is not performing unwanted casting.
        """
        reduce_funcs = [np.min, np.max, np.nanmin, np.nanmax]
        arr = FastArray(num_list)
        for dt in int_types + float_types:
            arr = arr.astype(dt)
            for f in reduce_funcs:
                reduced = np.array(f(arr)).dtype
                self.assertEqual(reduced, dt)

    # --------WARN ON INVALID-------------------
    def test_no_negatives(self):
        """
        FastArray does not warn on certain math functions that cannot take invalids.
        Numpy warns, but produces the same result.

        Make sure the result is NaN
        """
        arr = FastArray([-50])
        for f in no_negatives:
            # DJC: unfortunate handling of neg. value warning for log2.
            if f == np.log2:
                with self.assertWarns(RuntimeWarning):
                    b = f(arr)
            else:
                b = f(arr)
            b = isnan(b)
            self.assertTrue(np.all(b))

    # --------FLIPPING DIMENSIONS-----------------
    def test_1_dimension(self):
        """
        FastArray flips scalar initialization to 1 dimension (if the scalar is a supported type)
        Numpy creates an array of 0 dimensions

        Reduce functions on numpy arrays of 0 dimensions will produce a scalar.
        """
        dim0 = FastArray(1).ndim
        dim1 = FastArray([1]).ndim
        self.assertEqual(dim0, dim1)

    # --------PASS SQUEEZE RESULT OF 0 DIMENSIONS TO NUMPY----
    def test_squeeze_zero(self):
        """
        Because it does not support 0 dimensional arrays, FastArray will
        flip a 0 dimensional result of a squeeze() operation to Numpy.
        """
        arr = FastArray([1]).squeeze()
        self.assertFalse(isinstance(arr, FastArray))

    # ----------ATTEMPT TO CONVERT STRINGS-------
    def test_string_convert(self):
        """
        FastArray attempts to convert unicode strings to byte strings. If it can't, it will warn the user.
        """
        u = "AAPL\u2080"
        b = "AAPL"
        # with self.assertWarns(UserWarning):
        #    u_arr = FastArray(u)

        b_arr = FastArray([b])
        x = isinstance(b_arr[0], bytes)
        self.assertTrue(x)

    # --------TOO MANY DIMENSIONS > 1-----------
    def test_invalid_dims(self):
        """
        FastArray only handles one dimension correctly, so it warns the user when the input is more than one dimension.
        """
        multi_dims = np.empty((1, 20, 1, 40, 1))
        with self.assertWarns(UserWarning):
            _ = FastArray(multi_dims)

    # -------BINARY UFUNCS WITH SCALAR-------------------
    def test_binary_ufuncs_scalar(self):
        """
        Binary math ufuncs between numeric types and scalars must match numpy exactly.
        In certain cases however, FastArray returns a different dtype, so the CRC check will fail.

        In floor_divide, numpy returns -0.0 for floats: ex. 0.0/1
        FastArray returns 0.0
        """
        nonzero_nums = list(range(1, 10))

        num_types = int_types + float_types
        math_ufuncs = [
            np.add,
            np.subtract,
            np.multiply,
            np.divide,
            np.true_divide,
            np.floor_divide,
        ]
        for dt in num_types:
            fa_arr = FastArray(nonzero_nums, dtype=dt)
            np_arr = np.array(nonzero_nums, dtype=dt)
            for func in math_ufuncs:
                for scalar in (3, 2.1):
                    fa_result = func(fa_arr, scalar)
                    np_result = func(np_arr, scalar)
                    if fa_result.dtype == np_result.dtype:
                        fa_result = rc.CalculateCRC(fa_result)
                        np_result = rc.CalculateCRC(np_result)
                        self.assertEqual(
                            fa_result,
                            np_result,
                            msg=f"Test failed on function {func} with dtype {dt} and scalar {scalar}.",
                        )

    # -------BINARY UFUNCS WITH VECTOR-------------------
    def test_binary_ufuncs_vector(self):
        """
        Binary math ufuncs between numeric types and scalars must match numpy exactly.
        In certain cases however, FastArray returns a different dtype, so the CRC check will fail.

        This mixes and matches every numeric datatype and performs every hooked binary ufunc. If the result dtypes match, a CRC check will
        be performed.
        """

        num_types = int_types + float_types
        math_ufuncs = [func for func, v in gBinaryUFuncs.items() if v is not None]

        for dt1 in num_types:
            random_nums1 = (np.random.rand(10) * 100) + 1
            fa_nums1 = FastArray(random_nums1, dtype=dt1)
            np_nums1 = fa_nums1._np

            for dt2 in num_types:
                random_nums2 = (np.random.rand(10) * 100) + 1
                fa_nums2 = FastArray(random_nums2)
                np_nums2 = fa_nums2._np

                for func in math_ufuncs:
                    fa_result = func(fa_nums1, fa_nums2)
                    np_result = func(np_nums1, np_nums2)
                    if fa_result.dtype == np_result.dtype:
                        self.assertTrue(
                            almost_eq(fa_result, np_result, places=5),
                            msg=f"Test failed on function {func} with dtypes {fa_nums1.dtype} {fa_nums2.dtype}\n{fa_nums1}\n{fa_nums2}\n{fa_result}\n{np_result}",
                        )
                        # if func == np.power:
                        #    for i, _ in enumerate(fa_result):
                        #        fsub = np.log(fa_result[i]) - np.log(np_result[i])
                        #        self.assertAlmostEqual(fsub, 0, msg=f"Test failed on function {func} with dtypes {fa_nums1.dtype} {fa_nums2.dtype}\n{fa_nums1[i]}\n{fa_nums2[i]}\n{fa_result[i]}\n{np_result[i]}")
                        # else:

                        #    fa_resultcrc = rc.CalculateCRC(fa_result)
                        #    np_resultcrc = rc.CalculateCRC(np_result)
                        #    self.assertEqual(fa_resultcrc, np_resultcrc, msg=f"Test failed on function {func} with dtypes {fa_nums1.dtype} {fa_nums2.dtype}\n{fa_nums1}\n{fa_nums2}\n{fa_result}\n{np_result}")

    # ---------------UNARY UFUNCS-------------------
    def test_unary_ufuncs_scalar(self):
        """
        ***incomplete test - need to handle the rest of the unary funcs for as many types as possible

        np.negative and np.positive both raise errors when applied to boolean arrays
        """
        random_nums1 = (np.random.rand(10) * 100) + 1

        num_types = int_types + float_types
        # TODO: add separate test for boolean arrays to prevent natural crash
        # num_types.remove(np.bool_)
        math_ufuncs = [
            np.absolute,
            np.abs,
            np.fabs,
            np.floor,
            np.ceil,
            np.trunc,
            np.round,
            np.rint,
        ]

        for dt in num_types:
            fa_arr = FastArray(random_nums1, dtype=dt)
            np_arr = np.array(random_nums1, dtype=dt)
            for func in math_ufuncs:
                fa_result = func(fa_arr)
                np_result = func(np_arr)
                if fa_result.dtype == np_result.dtype:
                    fa_result = rc.CalculateCRC(fa_result)
                    np_result = rc.CalculateCRC(np_result)
                    self.assertEqual(
                        fa_result,
                        np_result,
                        msg=f"Test failed on function {func} with dtype {dt}",
                    )
                else:
                    self.assertAlmostEqual(
                        fa_result,
                        np_result,
                        msg=f"Test failed on function {func} with dtype {dt}",
                    )

    # ----------------TEST BITWISE -------------------------
    def test_bitwise(self):
        """
        These functions only apply to integer or boolean arrays.
        """
        bitwise_funcs = [np.bitwise_and, np.bitwise_xor, np.bitwise_or]
        for dt in int_types:
            fa_arr = FastArray(num_list, dtype=dt)
            np_arr = fa_arr._np
            for func in bitwise_funcs:
                fa_result = func(fa_arr, fa_arr[::-1])
                np_result = func(np_arr, np_arr[::-1])
                if fa_result.dtype == np_result.dtype:
                    fa_result = rc.CalculateCRC(fa_result)
                    np_result = rc.CalculateCRC(np_result)
                    self.assertEqual(
                        fa_result,
                        np_result,
                        msg=f"Test failed on function {func} with dtype {dt}",
                    )

    # ----------------TEST LOGICAL -------------------------
    def test_compare(self):
        """
        Compares FastArray results to numpy results for binary comparison and logical ufuncs.
        All results will be boolean arrays.

        There is a difference between calling a numpy ufunc and using a comparison operator,
        so the operators need to be checked separately.
        """
        basic_types = [np.int32, np.int64, np.float32, np.float64]
        numeric_types = int_types + float_types
        comparison_ufuncs = [
            np.less_equal,
            np.less,
            np.equal,
            np.not_equal,
            np.greater,
            np.greater_equal,
        ]
        logical_ufuncs = [np.logical_and, np.logical_xor, np.logical_or]
        comparison_operators = [
            "__ne__",
            "__eq__",
            "__ge__",
            "__gt__",
            "__le__",
            "__lt__",
        ]
        all_funcs = comparison_ufuncs + logical_ufuncs

        for dt1 in numeric_types:
            for dt2 in numeric_types:
                fa_arr1 = FA(num_list, dtype=dt1)
                fa_arr2 = FA(list(reversed(num_list)), dtype=dt2)
                np_arr1 = np.array(num_list, dtype=dt1)
                np_arr2 = np.array(list(reversed(num_list)), dtype=dt2)
                for func in all_funcs:
                    fa_result = func(fa_arr1, fa_arr2)
                    np_result = func(np_arr1, np_arr2)
                    # check that result lengths are the same
                    self.assertEqual(
                        len(fa_result),
                        len(np_result),
                        msg=f"Result sizes did not match for {func} with dtypes {dt1} {dt2}",
                    )
                    # compare each result item
                    arr_size = len(fa_result)
                    for i in range(arr_size):
                        self.assertEqual(
                            fa_result[i],
                            np_result[i],
                            msg=f"Comparison result did not match for {func} with dtypes {dt1} {dt2}",
                        )

                for f_name in comparison_operators:
                    fa_func = fa_arr1.__getattribute__(f_name)
                    np_func = np_arr1.__getattribute__(f_name)
                    fa_result = fa_func(fa_arr2)
                    np_result = np_func(np_arr2)
                    # check that result lengths are the same
                    self.assertEqual(
                        len(fa_result),
                        len(np_result),
                        msg=f"Result sizes did not match for operator {f_name} with dtypes {dt1} {dt2}",
                    )
                    # compare each result item
                    arr_size = len(fa_result)
                    for i in range(arr_size):
                        self.assertEqual(
                            fa_result[i],
                            np_result[i],
                            msg=f"Comparison operator {f_name} failed with dtypes {dt1} {dt2}",
                        )

    # -------------------------------------------------------------
    def test_np_vs_member(self):
        import builtins

        """
        Check to make sure the result is the same no matter how the ufunc is accessed.
        """
        func_names = ["min", "max", "sum"]
        arr = FA([1, 2, 3, 4, 5])
        num_types = int_types[1:] + float_types
        for dt in num_types:
            arr = arr.astype(dt)
            # print(dt)
            for name in func_names:
                member_func = None
                np_func = None
                builtin_func = None

                results = []
                member_func = arr.__getattribute__(name)
                results.append(member_func())
                if hasattr(np, name):
                    np_func = np.__getattribute__(name)
                    results.append(np_func(arr))
                if hasattr(builtins, name):
                    builtin_func = builtins.__getattribute__(name)
                    results.append(builtin_func(arr))

                self.assertEqual(
                    len(set(results)),
                    1,
                    msg=f"Results did not match for datatype {dt} and function {name}. Fastarray: {member_func} Numpy: {np_func} Builtin: {builtin_func}",
                )

                # we return different scalars than numpy, so the test below will fail
                # result_types = [ type(r) for r in results]
                # self.assertEqual(len(set(results)), 1, msg=f"Results did not match for datatype {dt} and function {name}. Fastarray: {member_func} Numpy: {np_func} Builtin: {builtin_func}" )

    # ---------------- UFUNCS ON EMPTY ----------------------------
    @pytest.mark.skip("Skipping until test is implemented.")
    def test_empty_ufuncs(self):
        n = np.array([])
        f = FA([])

        pass

    # ------------------------------------------------------
    def check_exception(self):
        """
        Example of checking that we raise the same error as numpy
        """
        try:
            n = np.array()
        except Exception as e:
            with self.assertRaises(e.__class__):
                f = FA()
        else:
            f = FA()
            self.assertEqual(
                f,
                n,
                msg="Empty constructor produced a different result for numpy and fastarray",
            )

    # ----------------TEST LOGICAL FOR STRING------------------------
    def test_string_compare(self):
        """
        FastArray currently does not support bytestring array comparison with ufuncs - numpy also prints notimplemented
        However operators <=, <, ==, !=, >, >= will return the correct result (boolean array)
        """
        f_arr = FA(["a", "b", "c"])
        invalid_funcs = [
            np.less_equal,
            np.less,
            np.equal,
            np.not_equal,
            np.greater,
            np.greater_equal,
        ]
        valid_func_names = ["__ne__", "__eq__", "__ge__", "__gt__", "__le__", "__lt__"]
        correct_results = [
            [False, False, False],
            [True, True, True],
            [True, True, True],
            [False, False, False],
            [True, True, True],
            [False, False, False],
        ]
        correct_dict = dict(zip(valid_func_names, correct_results))

        # ufunc comparisons will not work for strings (should we implement this on our own?)
        for func in invalid_funcs:
            with self.assertRaises(TypeError, msg=f"String comparison did not raise TypeError for {func}"):
                result = func(f_arr, f_arr)

        # strings need to be compared this way
        for f_name in valid_func_names:
            func = f_arr.__getattribute__(f_name)
            result = func(f_arr)
            correct = correct_dict[f_name]
            for i in range(len(result)):
                self.assertEqual(
                    result[i],
                    correct[i],
                    msg=f"String comparison failed for function {f_name}",
                )

    # ---------------ERROR FOR NP.POSITIVE, NP.NEGATIVE--------------
    def test_bool_bug1(self):
        """
        Note: FastArray does not hook these ufuncs right now, but if they are taken over,
        the same error should be raised when these functions are called.
        """
        arr = FastArray([True, False, True, False])

        with self.assertRaises(TypeError):
            _ = np.positive(arr)
        with self.assertRaises(TypeError):
            _ = np.negative(arr)

    # ------------------------------------------------------
    def test_display_properties(self):
        """
        FastArrays of default types have default item formatting for display (see Utils.rt_display_properties)
        This checks to see that the correct item format is being returned from a FastArray
        """
        f = FA(num_list, dtype=np.int32)
        item_format, convert_func = f.display_query_properties()
        self.assertEqual(
            item_format.length,
            DisplayLength.Short,
            msg=f"Incorrect length for item format.",
        )
        self.assertEqual(item_format.justification, DisplayJustification.Right)
        # self.assertEqual(item_format.invalid, None)
        self.assertEqual(item_format.can_have_spaces, False)
        self.assertEqual(item_format.color, DisplayColumnColors.Default)
        self.assertEqual(convert_func.__name__, "convertInt")

    # ------------------------------------------------------
    def test_ddof_default(self):
        """
        FastArray differs from numpy when calculating var, nanvar, std, and nanstd. We set ddof's default to 1 instead of 0
        """
        arr = FA([1, 2])
        func_names = ["std", "var", "nanstd", "nanvar"]
        for name in func_names:
            # make sure different than numpy
            func = arr.__getattribute__(name)
            fa_default = func()
            np_default = func(ddof=0)
            self.assertNotAlmostEqual(
                fa_default,
                np_default,
                msg=f"Failed to set ddof to 1 for reduce function {name}",
            )

            # make sure ddof is 1 when sent to numpy reduce
            sent_to_numpy = func(keepdims=0)
            self.assertAlmostEqual(
                fa_default,
                sent_to_numpy,
                msg=f"Failed to set ddof to 1 before passing to numpy for function {name}",
            )

    # ------------------------------------------------------
    def test_dtype_reduce(self):
        """
        If a dtype is passed to a reduce function, make sure the result is the correct dtype
        """
        dt = np.int8
        arr = FA([1, 2])
        func_names = ["std", "var", "nanstd", "nanvar"]
        for name in func_names:
            func = arr.__getattribute__(name)
            result = func(dtype=dt)
            self.assertEqual(
                dt,
                result.dtype,
                msg=f"Dtypes did not match for func {name}. {dt} was the keyword, the output was {result.dtype}",
            )

    # ------------------------------------------------------
    def test_dtype_int_reduce(self):
        """
        If the result of a reduce function is an integer, it will be cast to int64.
        If the input is uint64, the output will also be uint64
        """
        func_names = ["nansum"]
        unsigned_arr = FA(num_list, dtype=np.int32)
        signed_arr = FA(num_list, dtype=np.uint64)
        for name in func_names:
            us_func = unsigned_arr.__getattribute__(name)
            s_func = signed_arr.__getattribute__(name)
            us_dt = us_func().dtype
            s_dt = s_func().dtype
            self.assertEqual(us_dt, np.int64)
            self.assertEqual(s_dt, np.uint64)

    # ------------------------------------------------------
    def test_shift(self):
        """
        Check to make sure FastArray's shift mimics pandas shift - not numpy roll.
        """
        arr0 = FA([1, 2, 3, 4, 5])
        all_ints = int_types  # [1:] # bool is not included
        shift_dict = {
            tuple(float_types): np.nan,
            tuple([np.str_]): "",
            tuple([np.bytes_]): b"",
        }
        for tp in all_ints:
            if tp is bool:
                tp = np.bool_
            shift_dict[(tp,)] = INVALID_DICT[tp(1).dtype.num]
        for dtypes, invalid in shift_dict.items():
            for dt in dtypes:
                arr = arr0.astype(dt)
                pos_shift = arr.shift(1)
                neg_shift = arr.shift(-1)
                if invalid != invalid:
                    self.assertNotEqual(
                        pos_shift[0],
                        pos_shift[0],
                        msg=f"Positive shift on datatype {dt} did not fill with {invalid}.",
                    )
                    self.assertNotEqual(
                        neg_shift[-1],
                        neg_shift[-1],
                        msg=f"Negative shift on datatype {dt} did not fill with {invalid}.",
                    )
                else:
                    self.assertEqual(
                        pos_shift[0],
                        invalid,
                        msg=f"Positive shift on datatype {dt} did not fill with {invalid}.",
                    )
                    self.assertEqual(
                        neg_shift[-1],
                        invalid,
                        msg=f"Negative shift on datatype {dt} did not fill with {invalid}.",
                    )

                self.assertEqual(
                    pos_shift[1],
                    arr[0],
                    msg=f"Positive shift on datatype {dt} did not shift existing values to the correct place.",
                )
                self.assertEqual(
                    neg_shift[0],
                    arr[1],
                    msg=f"Negative shift on datatype {dt} did not shift existing values to the correct place.",
                )

    # FA supported
    def test_NotImplemented_not_returned(self):
        # See gh-5964 and gh-2091. Some of these functions are not operator
        # related and were fixed for other reasons in the past.

        # Cannot do string operation on
        binary_funcs = [
            np.power,
            np.subtract,
            np.multiply,
            np.divide,
            np.true_divide,
            np.floor_divide,
            np.bitwise_and,
            np.bitwise_or,
            np.bitwise_xor,
            np.left_shift,
            np.right_shift,
            np.fmax,
            np.fmin,
            np.fmod,
            np.hypot,
            np.logaddexp,
            np.logaddexp2,
            np.logical_and,
            np.logical_or,
            np.logical_xor,
            np.maximum,
            np.minimum,
            np.mod,
        ]

        # These functions still return NotImplemented. Will be fixed in
        # future.
        # bad = [np.greater, np.greater_equal, np.less, np.less_equal, np.not_equal]

        a = np.array("1")
        a = FA(a)
        b = 1
        for f in binary_funcs:
            with self.assertRaises(
                TypeError,
                msg=f"TypeError not raised for {f} with string and int input.",
            ):
                _ = f(a, b)


if __name__ == "__main__":
    tester = unittest.main()



from enum import IntEnum

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import riptable as rt
from riptable import (
    FA,
    Cat,
    Categorical,
    CatZero,
    Dataset,
    FastArray,
    arange,
    isnan,
    isnotnan,
    logical,
    mask_or,
)
from riptable.rt_enum import GROUPBY_KEY_PREFIX

#%load_ext autoreload
#%autoreload 2
# from riptable import *
# from enum import IntEnum
str_fa = FastArray(
    [
        "c",
        "e",
        "e",
        "d",
        "c",
        "b",
        "d",
        "c",
        "a",
        "b",
        "d",
        "e",
        "c",
        "a",
        "e",
        "d",
        "b",
        "a",
        "b",
        "c",
        "d",
        "b",
        "e",
        "c",
        "c",
        "d",
        "e",
        "c",
        "a",
        "c",
    ]
)
int_fa = FastArray(
    [
        10,
        20,
        10,
        20,
        10,
        10,
        20,
        10,
        20,
        20,
        30,
        20,
        20,
        20,
        20,
        30,
        10,
        10,
        10,
        20,
        30,
        20,
        20,
        10,
        30,
        30,
        20,
        10,
        10,
        10,
    ]
)
flt_fa = FastArray(
    [
        88.97,
        50.05,
        36.16,
        63.22,
        27.33,
        17.98,
        23.35,
        52.60,
        26.82,
        78.75,
        34.20,
        49.35,
        19.50,
        29.91,
        74.96,
        94.41,
        27.95,
        56.65,
        38.08,
        16.51,
        36.81,
        80.14,
        53.64,
        49.55,
        65.82,
        85.28,
        61.68,
        72.85,
        91.71,
        61.12,
    ]
)
tens = FastArray([10] * 30)
ds = Dataset({"strings": str_fa.copy(), "ints": int_fa, "floats": flt_fa, "tens": tens})
gb = ds.gb("strings")
ds_nums = Dataset({"ints": int_fa, "floats": flt_fa, "tens": tens})
data_to_compare = ["ints", "floats", "tens"]
gbu = ds.gbu("strings")

gb_funcs_L1 = [
    "sum",
    "mean",
    "min",
    "max",
    "var",
    "std",
    "nansum",
    "nanmean",
    "nanmin",
    "nanmax",
    "nanvar",
    "nanstd",
]
gb_funcs_L2 = ["first", "last", "median", "mode", "nanmedian"]
gb_funcs_L3 = ["cumsum", "cumprod"]
all_gb_ops = gb_funcs_L1 + gb_funcs_L2 + gb_funcs_L3

all_gb_kwargs = {func: [{}] for func in all_gb_ops}
all_gb_kwargs["quantile"] = [
    {"q": 0.245},
    {"q": 0.0},
    {"q": 0.713},
    {"q": 0.6},
    {"q": 0.501},
    {"q": 0.500000},
]

even_filter = logical(arange(30) % 2)
d_filter = str_fa != b"d"

d_filter_results = [
    0,
    np.nan,
    np.nan,
    np.nan,
    np.nan,
    np.nan,
    0,
    np.nan,
    np.nan,
    np.nan,
    np.nan,
    np.nan,
]


complete_unique_cats = FastArray(["a", "b", "c", "d", "e"])
unsorted_unique_cats = FastArray(["b", "c", "e", "a", "d"])
complete_unique_ints = FastArray([10, 20, 30, 40, 50])
unsorted_unique_ints = FastArray([20, 30, 50, 10, 40])
complete_unique_flts = FastArray([10.0, 20.0, 30.0, 40.0, 50.0])

sorted_codes = FastArray(
    [
        2,
        4,
        4,
        3,
        2,
        1,
        3,
        2,
        0,
        1,
        3,
        4,
        2,
        0,
        4,
        3,
        1,
        0,
        1,
        2,
        3,
        1,
        4,
        2,
        2,
        3,
        4,
        2,
        0,
        2,
    ]
)
sorted_codes_matlab = FastArray(
    [
        3.0,
        5.0,
        5.0,
        4.0,
        3.0,
        2.0,
        4.0,
        3.0,
        1.0,
        2.0,
        4.0,
        5.0,
        3.0,
        1.0,
        5.0,
        4.0,
        2.0,
        1.0,
        2.0,
        3.0,
        4.0,
        2.0,
        5.0,
        3.0,
        3.0,
        4.0,
        5.0,
        3.0,
        1.0,
        3.0,
    ]
)
str_fa_with_invalid = FastArray(
    [
        "g",
        "e",
        "e",
        "d",
        "g",
        "b",
        "d",
        "g",
        "a",
        "b",
        "d",
        "e",
        "g",
        "a",
        "e",
        "d",
        "b",
        "a",
        "b",
        "g",
        "d",
        "b",
        "e",
        "g",
        "g",
        "d",
        "e",
        "g",
        "a",
        "g",
    ]
)

last_array = FastArray(
    [
        b"Bryant",
        b"Copeland",
        b"Copeland",
        b"Hill",
        b"Bryant",
        b"Smith",
        b"Hill",
        b"Bryant",
        b"Jones",
        b"Smith",
        b"Hill",
        b"Copeland",
        b"Bryant",
        b"Jones",
        b"Copeland",
        b"Hill",
        b"Smith",
        b"Jones",
        b"Smith",
        b"Bryant",
        b"Hill",
        b"Smith",
        b"Copeland",
        b"Bryant",
        b"Bryant",
        b"Hill",
        b"Copeland",
        b"Bryant",
        b"Jones",
        b"Bryant",
    ]
)

first_array = FastArray(
    [
        b"David",
        b"Susan",
        b"Susan",
        b"Paul",
        b"David",
        b"Charlie",
        b"Paul",
        b"David",
        b"Sam",
        b"Charlie",
        b"Paul",
        b"Susan",
        b"David",
        b"Sam",
        b"Susan",
        b"Paul",
        b"Charlie",
        b"Sam",
        b"Charlie",
        b"David",
        b"Paul",
        b"Charlie",
        b"Susan",
        b"David",
        b"David",
        b"Paul",
        b"Susan",
        b"David",
        b"Sam",
        b"David",
    ]
)


class str_enum(IntEnum):
    a = 0
    b = 1
    c = 2
    d = 3
    e = 4


class TestCategoricalGroupby:
    def test_friendly_name(self):
        pass

    def get_gb_results(self, op_name, gb, c, c_data, cols_to_match):
        gb_call = getattr(gb, op_name)
        list_gb_kwargs = all_gb_kwargs[op_name]
        match = True
        for gb_kwargs in list_gb_kwargs:
            gb_result = gb_call(**gb_kwargs)
            gb_cat_call = getattr(c, op_name)
            gb_cat_result = gb_cat_call(c_data, **gb_kwargs)
            # print('gb\n',gb_result)
            # print('cat\n',gb_cat_result)
            for col in cols_to_match:
                match = match and bool(np.all(gb_result[col] == gb_cat_result[col]))
        return match

    def get_filtered_bin_results(self, op_name, c, c_data, col_name, filter, bin_idx, correct):
        match = False
        func = getattr(c, op_name)
        # print('flt_fa',flt_fa.get_name())
        empty_bin_result = func(flt_fa, filter=filter)[col_name][bin_idx]

        # if not nan
        if empty_bin_result == empty_bin_result:
            if empty_bin_result == correct:
                match = True
        else:
            if np.isnan(correct):
                match = True
        return match

    def funnel_all_tests(self, c, ds_gb, constructor_name, sorted=True):
        # no filter
        for op_name in all_gb_ops:
            # print('OP NAME WAS',op_name)
            match = self.get_gb_results(op_name, ds_gb, c, ds_nums, data_to_compare)
            assert (
                match
            ), f"Categorical {c} constructed from: {constructor_name} Results did not match for {op_name} operation."

        # post filter
        if sorted is True:
            bin_idx = 3
        else:
            bin_idx = 2

        for op_name, correct in zip(gb_funcs_L1, d_filter_results):
            match = self.get_filtered_bin_results(op_name, c, flt_fa, "floats", d_filter, bin_idx, correct)
            assert (
                match
            ), f"Categorical constructed from: {constructor_name} Incorrect result for filtered bin during {op_name} operation. Expected {correct}"

    def construction_with_invalid(self, c, constructor_name):
        for op_name, correct in zip(gb_funcs_L1, d_filter_results):
            match = self.get_filtered_bin_results(op_name, c, flt_fa, "floats", None, 2, correct)
            assert (
                match
            ), f"Categorical constructed from: {constructor_name} Incorrect result for filtered bin during {op_name} operation. Expected {correct}"

    def simple_string_set_item(self, *args, **kwargs):
        """
        This test needs to be updated with different data that reflects the new comparison behavior.
        SJK: 9/24/2018

        """

        source = kwargs["constructor_name"]
        del kwargs["constructor_name"]

        if "categories" in kwargs:
            kwargs["categories"] = kwargs["categories"].copy()

        c = Categorical(*args, **kwargs)
        set_items = [
            # index by string
            (b"b", b"a"),
            (b"b", "a"),
            # (b'b', 1),
            ("b", b"a"),
            ("b", "a"),
            # ('b', 1),
            # index by bool array
            # boolean arrays can no longer be generated with these comparisons SJK 9/24/2018
            (c == b"b", b"a"),
            (c == b"b", "a"),
            # (c == b'b', 1),
            (c == "b", b"a"),
            (c == "b", "a"),
            # (c == 'b', 1),
            # (c == 2, b'a'),
            # (c == 2, 'a'),
            # (c == 2, 1),
            # integer index
            ([5, 9, 16, 18, 21], b"a"),
            ([5, 9, 16, 18, 21], "a"),
            # ([ 5,  9, 16, 18, 21], 1),
            ([5, 9, 16, 18, 21], b"a"),
            ([5, 9, 16, 18, 21], "a"),
            # ([ 5,  9, 16, 18, 21], 1),
            ([5, 9, 16, 18, 21], b"a"),
            ([5, 9, 16, 18, 21], "a"),
            # ([ 5,  9, 16, 18, 21], 1),
        ]
        # this test needs to get reworked
        # no longer produces the correct result for all types of categoricals because of == comparison behavior
        for items in set_items:
            c = Categorical(*args, **kwargs)
            goal = c == ["a", "b"]
            c[items[0]] = items[1]
            result = c == items[1]
            all_set = np.sum(goal == result)
            assert all_set == 30, f"did not set c[{items[0]}] to {items[1]} for categorical from {source}"

            none_left = np.sum(c == "b")
            assert none_left == 0, f"did not set c[{items[0]}] to {items[1]} for categorical from {source}"

    def mk_set_item(self, *args, **kwargs):
        source = kwargs["constructor_name"]
        del kwargs["constructor_name"]

        if "categories" in kwargs:
            print("copying categories")
            kwargs["categories"] = kwargs["categories"].copy()

        c = Categorical(*args, **kwargs)
        set_items = [
            # index by string
            ((b"b", b"b"), (b"a", b"a")),
            ((b"b", b"b"), ("a", "a")),
            ((b"b", b"b"), 5),
            (("b", "b"), (b"a", b"a")),
            (("b", "b"), ("a", "a")),
            (("b", "b"), 5),
            # index by bool array
            # (c == (b'b', b'b'), (b'a', b'a')),
            # (c == (b'b', b'b'), ('a', 'a')),
            # (c == (b'b', b'b'), 5),
            # (c == ('b', 'b'), (b'a', b'a')),
            # (c == ('b', 'b'), ('a', 'a')),
            # (c == ('b', 'b'), 5),
            # (c == 4, (b'a', b'a')),
            # (c == 4, ('a', 'a')),
            # (c == 4, 5),
            # integer index
            ([5, 9, 16, 18, 21], (b"a", b"a")),
            ([5, 9, 16, 18, 21], ("a", "a")),
            ([5, 9, 16, 18, 21], 5),
            ([5, 9, 16, 18, 21], (b"a", b"a")),
            ([5, 9, 16, 18, 21], ("a", "a")),
            ([5, 9, 16, 18, 21], 5),
            ([5, 9, 16, 18, 21], (b"a", b"a")),
            ([5, 9, 16, 18, 21], ("a", "a")),
            ([5, 9, 16, 18, 21], 5),
        ]

        for items in set_items:
            c = Categorical(*args, **kwargs)
            goal = mask_or([c == ("a", "a"), c == ("b", "b")])
            c[items[0]] = items[1]
            result = c == items[1]
            all_set = np.sum(goal == result)
            assert all_set == 30, f"did not set c[{items[0]}] to {items[1]} for categorical from {source}"

            none_left = np.sum(c == ("b", "b"))
            assert none_left == 0, f"did not set c[{items[0]}] to {items[1]} for categorical from {source}"

    # TODO pytest parameterize funnel_all_tests
    # --STRINGS-------------------------------------------------------------------------------
    def test_groupby_ops_string_list(self):
        c = Categorical(str_fa)

        self.funnel_all_tests(c, gb, "string list")
        self.simple_string_set_item(str_fa, constructor_name="string list")

    # --STRINGS + CATEGORIES------------------------------------------------------------------
    def test_groupby_ops_string_list_cats(self):
        c = Categorical(str_fa, complete_unique_cats)
        self.funnel_all_tests(c, gb, "string list + categories")

        self.simple_string_set_item(
            str_fa,
            categories=complete_unique_cats,
            constructor_name="string list + categories",
        )

        # 5/16/2019 invalid category must be in uniques
        # c_invalid = Categorical(str_fa_with_invalid, categories=complete_unique_cats, invalid='invalid')
        # self.construction_with_invalid(c_invalid, 'string list + categories')

    # --INDEX + CATEGORIES + BASE 0-----------------------------------------------------------
    def test_groupby_ops_user_codes_base_0(self):
        c = Categorical(sorted_codes.copy(), categories=complete_unique_cats, base_index=0)
        self.funnel_all_tests(c, gb, "index + categories + base_index 0")

        c = CatZero(sorted_codes.copy(), categories=complete_unique_cats)
        self.funnel_all_tests(c, gb, "index + categories + base_index 0")

        # self.simple_string_set_item(sorted_codes, categories=complete_unique_cats, base_index=0, constructor_name="string list")

    # --INDEX + CATEGORIES + BASE 1-----------------------------------------------------------
    def test_groupby_ops_user_codes_base_1(self):
        c = Categorical(sorted_codes + 1, complete_unique_cats, base_index=1)
        self.funnel_all_tests(c, gb, "index + categories + base_index 1")

        self.simple_string_set_item(
            sorted_codes.copy(),
            categories=complete_unique_cats,
            base_index=1,
            constructor_name="index + categories + base_index 1",
        )

    # --INDEX + INTEGER CATEGORIES + BASE 0-----------------------------------------------------------
    def test_groupby_ops_user_codes_int_base_0(self):
        c = Categorical(sorted_codes.copy(), complete_unique_ints, base_index=0)
        self.funnel_all_tests(c, gb, "index + integer categories + base_index 0")

        c = CatZero(sorted_codes.copy(), complete_unique_ints)
        self.funnel_all_tests(c, gb, "index + integer categories + base_index 0")

    # --INDEX + INTEGER CATEGORIES + BASE 1-----------------------------------------------------------
    def test_groupby_ops_user_codes_int_base_1(self):
        c = Categorical(sorted_codes + 1, complete_unique_ints, base_index=1)
        self.funnel_all_tests(c, gb, "index + integer categories + base_index 1")

    # --INDEX + FLOAT CATEGORIES + BASE 0-----------------------------------------------------------
    def test_groupby_ops_user_codes_float_base_0(self):
        c = Categorical(sorted_codes.copy(), complete_unique_flts, base_index=0)
        self.funnel_all_tests(c, gb, "index + integer categories + base_index 0")

        c = CatZero(sorted_codes.copy(), complete_unique_flts)
        self.funnel_all_tests(c, gb, "index + integer categories + base_index 0")

    # --INDEX + FLOAT CATEGORIES + BASE 1-----------------------------------------------------------
    def test_groupby_ops_user_codes_float_base_1(self):
        c = Categorical(sorted_codes + 1, complete_unique_flts, base_index=1)

    # --INDEX + INTENUM------------------------------------------------------------------------
    def test_groupby_ops_int_enum(self):
        c = Categorical(sorted_codes, str_enum)
        self.funnel_all_tests(c, gbu, "index + int enum", sorted=False)

        # self.simple_string_set_item(sorted_codes, str_enum, constructor_name="string list")

    # --INDEX + MAPPING------------------------------------------------------------------------
    def test_groupby_ops_mapping(self):
        d = dict(str_enum.__members__)
        d = {k: int(v) for k, v in d.items()}
        c = Categorical(sorted_codes, d)
        self.funnel_all_tests(c, gbu, "index + mapping dictionary", sorted=False)

    # --PANDAS CATEGORIAL----------------------------------------------------------------------
    def test_groupby_ops_pd_cat(self):
        import pandas as pd

        pdc = pd.Categorical(str_fa)
        c = Categorical(pdc)
        self.funnel_all_tests(c, gb, "pandas categorical")

        # need to copy provided categories for an accurate test
        # self.simple_string_set_item(pdc, constructor_name="pandas categorical")

    # --MATLAB CATEGORIAL----------------------------------------------------------------------
    def test_groupby_ops_matlab(self):
        c = Categorical(sorted_codes_matlab, complete_unique_cats, from_matlab=True)
        self.funnel_all_tests(c, gb, "matlab categorical")

    # --STRINGS + UNSORTED CATEGORIES-----------------------------------------------------------
    def test_groupby_ops_string_list_cats_unsorted(self):
        c = Categorical(str_fa, unsorted_unique_cats, ordered=False, sort_gb=True)
        self.funnel_all_tests(c, gb, "string list + unsorted categories")

    # --MULTIKEY DICTIONARY---------------------------------------------------------------------
    def test_groupby_ops_multikey_dict(self):
        mk_dict = {"string1": str_fa, "string2": str_fa}
        mk_gb = Dataset(
            {
                "string1": str_fa.copy(),
                "string2": str_fa.copy(),
                "ints": int_fa,
                "floats": flt_fa,
                "tens": tens,
            }
        ).gbu(["string1", "string2"])
        c = Categorical(mk_dict)
        self.funnel_all_tests(c, mk_gb, "multikey dictionary", sorted=False)

        # setitem hits comparison functions - need to rewrite these tests after comparison behavior change
        # self.mk_set_item(mk_dict, constructor_name="multikey dictionary")

        # conflicting names
        x = str_fa.copy()
        y = str_fa.copy()
        z = str_fa.copy()
        x.set_name("strings")
        y.set_name("strings")
        z.set_name("strings1")
        c = Categorical([x, y, z])
        assert (
            c._categories_wrap.ncols == 3
        ), f"incorrect number of columns for multikey from list. {c._categories_wrap.ncols} vs. 3"
        # 04/25/2019 all default column names now happen in grouping object
        assert list(c.categories().keys()) == [
            "strings",
            GROUPBY_KEY_PREFIX + "_c1",
            "strings1",
        ], f"column names did not match for multikey from list. {list(c.categories().keys())} vs. ['strings','strings2','strings1']"

    # --MULTIKEY LIST---------------------------------------------------------------------------
    def test_groupby_ops_multikey_list(self):
        mk_list = [str_fa.copy(), str_fa.copy()]
        mk_gb = Dataset(
            {
                "string1": str_fa.copy(),
                "string2": str_fa.copy(),
                "ints": int_fa,
                "floats": flt_fa,
                "tens": tens,
            }
        ).gbu(["string1", "string2"])
        c = Categorical(mk_list)
        self.funnel_all_tests(c, mk_gb, "multikey list", sorted=False)

        # setitem hits comparison functions - need to rewrite these tests after comparison behavior change
        # self.mk_set_item(mk_list, constructor_name="multikey list")

    # --MULTIKEY COUNT--------------------------------------------------------------------------
    def test_multikey_count(self):
        mk_list = [str_fa.copy(), int_fa.copy(), str_fa.copy(), int_fa.copy()]
        c_multi = Categorical(mk_list)
        result_counts = c_multi.count().Count
        correct_counts = FastArray([6, 5, 1, 2, 3, 2, 2, 4, 2, 2, 1])
        all_correct = bool(np.all(result_counts == correct_counts))
        assert all_correct, f"Incorrect result for multikey count for 4 keys. {result_counts} vs. {correct_counts}"

    # --STRING-LIKE SINGLE KEY COUNT------------------------------------------------------------
    def test_single_key_string_count(self):
        correct_counts = FastArray([4, 5, 9, 6, 6])

        # for sorting/count bug fix 8/21/2018
        c_make_unique = Categorical(str_fa)
        result_counts = c_make_unique.count().Count
        match = bool(np.all(result_counts == correct_counts))
        assert match

        c_from_codes = Categorical(sorted_codes, complete_unique_cats, base_index=0)
        result_counts = c_from_codes.count().Count
        match = bool(np.all(result_counts == correct_counts))
        assert match

        c_from_codes_unsorted = Categorical(sorted_codes, unsorted_unique_cats, base_index=0)
        result_counts = c_from_codes_unsorted.count().Count
        match = bool(np.all(result_counts == correct_counts))
        assert match
        # 8/24/2018 SJK - default name for groupby key columns might change, so selected this by index
        # also, in most cases (save intenum/dict) categorical groupby no longer returns a categorical
        result_keys = c_from_codes_unsorted.count()[1]
        match = bool(np.all(result_keys == unsorted_unique_cats))
        assert match, f"Result: {result_keys} Expected: {unsorted_unique_cats}"

    def test_cumcount_vs_gb(self):
        arr = np.random.choice(["a", "b", "c", "d", "e"], 50)
        ds = Dataset({"keycol": arr, "col1": arange(50), "col2": arange(50)})
        gb_result = ds.gb("keycol").cumcount()

        c = Categorical(ds.keycol)
        c_result = c.cumcount()

        rdiff = gb_result - c_result
        assert sum(rdiff) == 0

        f = logical(arange(50) % 2)
        c_result = c.cumcount(filter=f)
        assert bool(np.all(isnotnan(c_result[f])))
        assert bool(np.all(isnan(c_result[~f])))

    # --MULTIKEY SINGLE STRING KEY--------------------------------------------------------------
    def test_groupby_ops_multikey_single_string(self):
        c = Categorical({"string_col": str_fa.copy()})
        self.funnel_all_tests(c, gb, "multikey single string key")

    # --MULTIKEY SINGLE NUMERIC KEY-------------------------------------------------------------
    def test_groupby_ops_multikey_single_numeric(self):
        c = Categorical({"codes": sorted_codes}, ordered=False, sort_gb=False)
        self.funnel_all_tests(c, gbu, "multikey single numeric key", sorted=False)

    # TODO pytest parameterize empty_results table
    def test_empty_category(self):
        # 5/16/2019 invalid category must be in uniques
        # c = Categorical(str_fa_with_invalid, complete_unique_cats, invalid='invalid')
        # can test empty bin like this, the third result will be empty
        c = Categorical(np.random.choice(["a", "b", "d", "e"], 30), ["a", "b", "c", "d", "e"])
        empty_result = [
            ("sum", 0.0),
            ("mean", np.nan),
            ("min", np.nan),
            ("max", np.nan),
            ("var", np.nan),
            ("std", np.nan),
            ("nansum", 0.0),
            ("nanmean", np.nan),
            ("nanmin", np.nan),
            ("nanmax", np.nan),
            ("nanvar", np.nan),
            ("nanstd", np.nan),
        ]

        for correct_tup in empty_result:
            func = getattr(c, correct_tup[0])
            result = func(ds_nums).floats[2]
            a = np.isnan(correct_tup[1])

            if np.isnan(correct_tup[1]):
                assert (
                    result != result
                ), f"Did not product correct result for empty category after {correct_tup[0]} operation."
            else:
                assert (
                    result == correct_tup[1]
                ), f"Did not product correct result for empty category after {correct_tup[0]} operation."

    def test_igroup_dtype(self):
        """
        The categorical only has 3 categories, so its index array will be int8
        after pack_by_group() is called in a grouping operation, it generates iGroup - which
        needs to be large enough to hold an integer equal to the length of the categorical
        """
        c = Categorical(np.random.choice([b"a", b"b", b"c"], 1_000_000))
        c_was_int8 = np.int8 == c.dtype
        assert c_was_int8

        _ = c.groups
        igroup_was_int32 = c.grouping.iGroup.dtype == np.int32
        assert igroup_was_int32

    # --TEST ALL CONSTRUCTOR FLAGS-----------------------------------------------------------
    def test_pre_filter(self):
        c = Categorical(str_fa, filter=even_filter)
        assert c._filter == None

        result = c.sum(ds_nums)
        one_fifty = sum(result.tens)
        assert one_fifty == 150

    def test_specify_gb_data(self):
        str_col = ["a", "a", "b", "c", "a"]
        num_col = [10, 10, 20, 30, 10]
        col1 = np.arange(5)
        col2 = np.arange(5)
        small_ds = Dataset({"str_col": str_col, "num_col": num_col, "col1": col1, "col2": col2})
        ds_to_operate_on = small_ds[["col1", "col2"]]

        c = Categorical(str_col)

        # dataset
        d = c.sum(ds_to_operate_on)

        # single
        # list
        d = c.sum([col1, col2])

        # tuple
        d = c.sum((col1, col2))

        # dict
        d = c.sum({"a": col1, "b": col2})

        # multiple
        d = c.sum(col1, col2)

    def test_as_categorical(self):
        ds = Dataset(
            {
                "keycol1": np.random.choice(["a", "b", "c"], 30),
                "keycol2": np.random.choice(["a", "b", "c"], 30),
                "data": np.random.rand(30),
            }
        )

        gbu = ds.gbu("keycol1")
        c = Categorical(ds.keycol1, ordered=False, sort_gb=False)
        cgbu = gbu.as_categorical()

        gbu_result = gbu.sum()
        c_result = c.sum(ds.data)
        cgbu_result = cgbu.sum(ds.data)

        for name, col in gbu_result.items():
            assert bool(np.all(c_result[name] == col))
            assert bool(np.all(cgbu_result[name] == col))

    def test_gb_labels_enum(self):
        # make sure enum groupby keys are displayed as string,  not integer code
        c = Categorical([10, 10, 10, 20, 30, 20, 10, 20, 20], {"a": 30, "b": 20, "c": 10})
        c_result = c.count()
        c_labels = c_result[c_result.label_get_names()][0]

        ds = Dataset({"catcol": c, "data": arange(9)})
        ds_result = ds.gbu("catcol").count()
        ds_labels = ds_result[ds_result.label_get_names()][0]

        assert c_labels.dtype.char == ds_labels.dtype.char
        assert bool(np.all(c_labels == ds_labels))

    def test_groupby_categorical_sort(self):
        """
        Test that groupby on a categorical sorts the dataset correctly
        """
        ds = rt.Dataset()
        cats = ["z", "y", "x", "w", "a", "b", "c", "d"]
        vals = [0, 1, 2, 3, 4, 5, 6, 7]
        expected = dict(zip(cats, vals))

        ds["Cat"] = rt.Categorical([cats[xx % len(cats)] for xx in range(100)])

        # two identical columns
        ds["Value1"] = [vals[xx % len(cats)] for xx in range(100)]
        ds["Value2"] = [vals[xx % len(cats)] for xx in range(100)]

        grp = ds.groupby("Cat").mean()
        grp["Expected"] = [expected[xx] for xx in grp.Cat.astype("U")]

        diff = rt.sum(rt.abs(grp.Expected - grp.Value1))
        diff += rt.sum(rt.abs(grp.Expected - grp.Value2))

        assert diff <= 1e-9

    def test_shift(self):
        """
        Test that Categorical.shift shifts the values in an array or Dataset *per group*.
        """
        result = rt.Cat([1, 1, 1, 2]).shift(arange(4), window=1)[0]
        assert result[1] == 0
        assert result[2] == 1

        result = rt.Cat([1, 1, 1, 2]).shift([5, 6, 7, 8], window=1)[0]
        assert result[1] == 5
        assert result[2] == 6

    def test_fill_backward(self):
        """
        Test that Categorical.fill_backward fills values backward *per group*.
        """
        data = rt.FA([1.0, 4.0, np.nan, np.nan, 9.0, 16.0])
        cat = rt.Categorical(["A", "B", "A", "B", "A", "B"])

        result = cat.fill_backward(data)

        # The result of this function should be a Dataset.
        assert isinstance(result, rt.Dataset)

        # The dataset should have the same number of rows as the data arrays
        # we operated on (an invariant of apply_nonreduce/"scan"/"prefix sum").
        assert result.shape[0] == len(data)

        # The dataset should have (N+M) columns, where N is the number
        # of keys within the Categorical and M is the number of columns
        # we performed the operation on.
        expected_col_count = len(cat.category_dict) + 1
        assert result.shape[1] == expected_col_count

        # Check the resulting data; the dtype of the data should be the
        # same as the original column.
        assert_array_equal(result[0], rt.FA([1.0, 4.0, 9.0, 16.0, 9.0, 16.0]))

    def test_fill_forward(self):
        """
        Test that Categorical.fill_forward fills values forward *per group*.
        """
        data = rt.FA([1.0, 4.0, 9.0, 16.0, np.nan, np.nan])
        cat = rt.Categorical(["A", "B", "A", "B", "A", "B"])

        result = cat.fill_forward(data)

        # The result of this function should be a Dataset.
        assert isinstance(result, rt.Dataset)

        # The dataset should have the same number of rows as the data arrays
        # we operated on (an invariant of apply_nonreduce/"scan"/"prefix sum").
        assert result.shape[0] == len(data)

        # The dataset should have (N+M) columns, where N is the number
        # of keys within the Categorical and M is the number of columns
        # we performed the operation on.
        expected_col_count = len(cat.category_dict) + 1
        assert result.shape[1] == expected_col_count

        # Check the resulting data; the dtype of the data should be the
        # same as the original column.
        assert_array_equal(result[0], rt.FA([1.0, 4.0, 9.0, 16.0, 9.0, 16.0]))

    def test_nobins(self):
        """
        Tests that Categorical.median() works correctly when there
        are no bins with data because the Categorical has been created
        with a pre-filter which has filtered out all data in the Dataset.
        """

        data = Dataset()
        data.Group = np.random.randint(0, 10, 100_000)
        data.Values = np.random.randint(0, 10, 100_000)
        x = data.cat("Group", filter=data.Group < 0)
        x.median(data.Values)

    def test_count(self):
        """
        Tests that Categorical.count(transform=True) works correctly.
        """

        x = Cat(np.arange(10) % 3)
        y = x.count(transform=True)[0]
        assert_array_equal(y, FA([4, 3, 3, 4, 3, 3, 4, 3, 3, 4]))

    def test_count_filter_transform(self):
        x = rt.Cat(rt.arange(100) % 10)
        f = rt.arange(100) % 4 != 0
        assert_array_equal(x.count(filter=f, transform=True).Count, x.nansum(f, transform=True).col_0)

        x = rt.Cat(str_fa)
        f = rt.arange(len(x)) % 3 != 0
        assert_array_equal(x.count(filter=f, transform=True).Count, x.nansum(f, transform=True).col_0)
        soln = rt.FA([5, 6, 6, 3, 5, 2, 3, 5, 4, 2, 3, 6, 5, 4, 6, 3, 2, 4, 2, 5, 3, 2, 6, 5, 5, 3, 6, 5, 4, 5])
        assert_array_equal(x.count(filter=f, transform=True).Count, soln)

    def test_length_error(self):
        x = rt.Cat(rt.arange(5))
        y = rt.arange(4)
        with pytest.raises(
            ValueError,
            match="Tried to perform a groupby operation where the length of the input was not equal to the length of the categorical.",
        ):
            _ = x.median(y)



"""
Functions for working with numba.

Functionality includes:

- Functions for numba <-> riptable interop
- Convenience functions and decorators used when implementing numba-accelerated functions.
"""
__all__ = []
